/*
==============================================================================
Title: "BSP Entity Extractor"
(c) 1999 by Matthias "Maddes" Buecher

Authors:
Matthias "Maddes" Buecher <maddes@bigfoot.com>

Most of this code is easy to comprehend and there should not be too many
questions about how it works.  The heart of the program has been slighty
commented out to help you understand the technical operations being done.

Important:
It should compile under Win32 and Unix with no errors, some warnings may still occur.
If you use a processor which has another byte order than Intel (lowest to highest byte),
you have to reprogram all offset functions, etc.
The file functions already handle DOS/Windows and Unix files correctly.

DISCLAIMER WARNING:
WE DO NOT ACCEPT RESPONSIBILITY FOR ANY EFFECTS, POSITIVE OR NEGATIVE,
THAT THIS CODE MAY CAUSE TOWARDS YOUR COMPUTER.
WE DO NOT SUPPORT MODIFICATIONS OF THIS CODE, USE ONLY AT YOUR OWN RISK.
THIS PATCH HAS NOT BEEN ENDORSED BY THE ID SOFTWARE CORPORATION.
THIS PROGRAM IS FREE OF CHARGE. IF YOU HAVE RECEIVED THIS PROGRAM THROUGH
PAYMENT, PLEASE CONTACT US.
THANK YOU.
==============================================================================
*/

/* Compatible headers */
#include <stddefs.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

/* needed to compile under Linux */
#ifndef __WIN32__
#define stricmp strcasecmp
#endif

/* constants */
#define VERSION		"1.00 (1999-11-29)"
#define PARM_CHAR	'-'
#define BUFLEN		1024

/*
structure definitions taken from Quake tool source codes
*/

/* BSP file */
#define BSP_Q1TVERSION	0x17	/* Classic Quake QTEST */
#define BSP_Q1VERSION	0x1d	/* Classic Quake */
#define BSP_IBSP_ID	'IBSP'	/* IBSP ID */
#define BSP_Q2VERSION	0x26	/* Quake II */
#define BSP_Q3VERSION	0x2d	/* Quake 3 Arena */

#define LUMP_ENTITIES		0
#define LUMP_PLANES		1
#define LUMP_TEXTURES		2
#define LUMP_VERTEXES		3
#define LUMP_VISIBILITY		4
#define LUMP_NODES		5
#define LUMP_TEXINFO		6
#define LUMP_FACES		7
#define LUMP_LIGHTING		8
#define LUMP_CLIPNODES		9
#define LUMP_LEAFS		10
#define LUMP_MARKSURFACES 	11
#define LUMP_EDGES		12
#define LUMP_SURFEDGES		13
#define LUMP_MODELS		14

#define	HEADER_LUMPS	15

typedef struct {
	long fileofs, filelen;		/* info about lump */
} lump_t;

typedef struct {			/* BSP file header */
	long version;			/* BSP file version */
	lump_t lumps[HEADER_LUMPS];	/* info table for lumps */
} dheader_t;

typedef struct {			/* BSP file header */
	long ibsp;			/* BSP file id */
	long version;			/* BSP file version */
	lump_t lumps[HEADER_LUMPS];	/* info table for lumps */
} dheader2_t;

/*
structure definitions for attribute list
*/
typedef struct s_attrentry {
	char line[BUFLEN+1];
	struct s_attrentry *next;
} attrentry;

/* global data definition */
char data_filename[BUFLEN+1];		/* data file name */
FILE *data;				/* data file handle */

char ent_filename[BUFLEN+1];		/* ent file name */
FILE *ent;				/* ent file handle */

char *extension;			/* pointer to extension string */

int flag_error;				/* error occured */
int flag_warning;			/* warning occured */

int flag_invert;			/* invert flag */

char line[BUFLEN+1];			/* line buffer */
char *line_char;			/* pointer to a char in line */
int linecount;				/* line counter */

char *entlist;				/* pointer to start of entity list */
char *entlist_char;			/* pointer to a char in entity list */


/*
******************************************************************************
*
*   G E N E R A L   F U N C T I O N S
*
******************************************************************************
*/

/*
==============================================================================
transforms a string to lowercase
==============================================================================
*/
void strlower(char *text)
{
	int i;
	int length;

	length = strlen(text);
	for (i=0; i<length; i++)
		text[i] = tolower(text[i]);
}

/*
==============================================================================
gets the extension out of the path
==============================================================================
*/
char *get_extension(const char *const filename)
{
	char *dot;
	char *slash;
	char *backslash;

	dot = strrchr(filename, '.');
	slash = strrchr(filename, '/');
	backslash = strrchr(filename, '\\');

	if ((dot) && (dot > slash) && (dot > backslash))
	{
		return dot;
	}

	return NULL;		// no extension found
}

/*
******************************************************************************
*
*   U S E R   I N T E R F A C E
*
******************************************************************************
*/

/*
==============================================================================
displays the help text
==============================================================================
*/
void display_help(char *call)
{
	/* syntax */
	printf("Usage: BSPEntEx [-?] [-i] <bspfile|mapfile> [<entfile>]\n");
	printf("  -?           : displays this help text\n");
	printf("  -i           : inverts the attribute order within entities\n");

	/* example */
	printf("\nExample:\n");
	printf("%s -i \"example.bsp\"\n", call);
	printf("This extracts the entity list with inverted attribute order to \"example.ent\"\n");

	/* contact */
	printf("\nHow to contact the author:\n");
	printf("Matthias \"Maddes\" Buecher <maddes@bigfoot.com>\n");
}

/*
==============================================================================
checks if string begins with a parameter character
==============================================================================
*/
int chk_paramchar(char *text)
{
	/* check for parameter character at beginning */
	if (text[0] == PARM_CHAR)
	{
		strlower(text);
		return 1;
	}
	else
		return 0;
}

/*
==============================================================================
checks all parameters
when help parameter is found it displays the help text and leaves the function
checks if the ini file is present
displays all used parameters and asks for confirmation
==============================================================================
*/
void chk_param(int argc, char *argv[])
{
	/* define local variables */
	int i;			/* iteration variable */
	int flag_help;		/* display help */

	/* initialize local variables */
	flag_error = 0;
	flag_warning = 0;
	flag_help = 0;

	/* parse parameters */
	for (i=1; i<argc; i++)
	{
		if (chk_paramchar(argv[i]))
		{
			if (!strcmp(argv[i],"-?") || !stricmp(argv[i],"-h") || !stricmp(argv[i],"-help"))	/* help parameter */
			{
				flag_help = 1;
				flag_error = 1;
				break;
			}
			else if (!stricmp(argv[i],"-i"))	/* invert parameter */
			{
				/* check already set */
				if (flag_invert)
				{
					printf("Warning on parameter %i: invert already defined\n", i);
					flag_warning = 1;
				}

				flag_invert=1;
			}
			else				/* unknown parameter */
			{
				printf("Error on parameter %i: \"%s\" is an unknown parameter\n", i, argv[i]);
				flag_help = 1;
				flag_error = 2;
			}
		}
		else	/* should be the filenames */
		{
			if (!*data_filename)
			{
				strcpy(data_filename, argv[i]);
			}
			else if (!*ent_filename)
			{
				strcpy(ent_filename, argv[i]);
			}
			else
			{
				printf("Error on parameter %i: \"%s\", a data and ent file were already named\n", i, argv[i]);
				flag_help = 1;
				flag_error = 2;
			}
		}
	}

	/* check if bsp file exists */
	if (!*data_filename)
	{
		printf("Error: no data file named\n");
		flag_error = 2;
		flag_help = 1;
	}
	else
	{
		data = fopen(data_filename, "rb");
		if (!data)
		{
			// check extension
			extension = get_extension(data_filename);
			if ( (!extension) || (stricmp(extension, ".bsp")) || (stricmp(extension, ".map")))
			{
				strcat(data_filename, ".bsp");
				data = fopen(data_filename, "rb");
			}
		}

		if (!data)
		{
			printf("Error: could not open data file \"%s\" - %s\n", data_filename, strerror(errno));
			flag_error = 2;
		}
		else
		{
			fclose(data);
			data = NULL;
		}
	}

	/* check ent filename */
	if (!*ent_filename)
	{
		strcpy(ent_filename, data_filename);
	}
	extension = get_extension(ent_filename);
	if ( (!extension) || (stricmp(extension, ".bsp")) )
	{
		strcat(ent_filename, ".ent");
	}
	else
	{
		strcpy(extension, ".ent");
	}

	/* asked for help */
	if (flag_error > 1 || flag_warning)
	{
		printf("\n");
	}

	if (flag_help)
	{
		display_help(argv[0]);
	}

	if (!flag_error)
	{
		printf("Data file: \"%s\"\n", data_filename);
		printf("ENT file: \"%s\"\n", ent_filename);

		if (flag_invert) printf("Inverted attribute order enabled\n");

		printf("\n");
	}
}

/*
******************************************************************************
*
*   M A P   F I L E   P A R S I N G
*
******************************************************************************
*/

int get_line(void)
{
	int linelength;				/* line length */

	linelength = 0;
	line_char = line;

	linecount++;

	while ( *entlist_char )
	{
		if (linelength == BUFLEN)
		{
			printf("Line %i exceeds maximum line buffer of %i", linecount, BUFLEN);
			flag_error = 2;
			break;
		}
		else
		{
			linelength++;
			*line_char++ = *entlist_char;

			if (*entlist_char++ == '\n')
			{
				break;
			}
		}
	}

	*line_char = 0;

	return linelength;
}

void get_firstchar(void)
{
	line_char = line;

	while ( *line_char==' ' || *line_char=='\t' )
	{
		line_char++;
	}
}

/*
******************************************************************************
*
*   F I L E S   R E A D I N G   &   W R I T I N G
*
******************************************************************************
*/

void process_file(void)
{
	dheader_t bspheader;
	int level;
	long filelength;
	long fileoffset;

	attrentry *attrlist;
	attrentry *attribute;

	/* initialize local variables */
	entlist = NULL;

	do			/* just to use break for errors */
	{
		extension = get_extension(data_filename);
		if ( (!extension) || (stricmp(extension, ".map")))
		{
			/* open bsp file */
			data = fopen(data_filename, "rb");
			if (!data)
			{
				printf("Error: could not open bsp file \"%s\" - %s\n", data_filename, strerror(errno));
				flag_error = 2;
				break;
			}

			/* read bsp header */
			if ( fseek(data, 0, SEEK_SET) != 0 || ftell(data) != 0 )
			{
				printf("Error: couldn't seek bsp header - %s\n", strerror(errno));
				flag_error = 2;
				break;
			}
			if ( fread(&bspheader, sizeof(dheader_t), 1, data) != 1 )
			{
				printf("Error: couldn't read bsp header - %s\n", strerror(errno));
				flag_error = 2;
				break;
			}

			/* checking bsp version / fileformat */
			if ( bspheader.version == (long)BSP_Q1TVERSION )
			{
				printf("Detected QTest BSP version %li\n", bspheader.version);
				fileoffset = bspheader.lumps[LUMP_ENTITIES].fileofs;
				filelength = bspheader.lumps[LUMP_ENTITIES].filelen;
			}
			else if (bspheader.version == (long)BSP_Q1VERSION )
			{
				printf("Detected Quake BSP version %li\n", bspheader.version);
				fileoffset = bspheader.lumps[LUMP_ENTITIES].fileofs;
				filelength = bspheader.lumps[LUMP_ENTITIES].filelen;
			}
			else if ( ((dheader2_t *)&bspheader)->ibsp == (long)BSP_IBSP_ID
			       && ((dheader2_t *)&bspheader)->version == (long)BSP_Q2VERSION )
			{
				printf("Detected Quake II BSP version %li\n", ((dheader2_t *)&bspheader)->version);
				fileoffset = ((dheader2_t *)&bspheader)->lumps[LUMP_ENTITIES].fileofs;
				filelength = ((dheader2_t *)&bspheader)->lumps[LUMP_ENTITIES].filelen;
			}
			else if ( ((dheader2_t *)&bspheader)->ibsp == (long)BSP_IBSP_ID
			       && ((dheader2_t *)&bspheader)->version == (long)BSP_Q3VERSION )
			{
				printf("Detected Quake 3 Arena BSP version %li\n", ((dheader2_t *)&bspheader)->version);
				fileoffset = ((dheader2_t *)&bspheader)->lumps[LUMP_ENTITIES].fileofs;
				filelength = ((dheader2_t *)&bspheader)->lumps[LUMP_ENTITIES].filelen;
			}
			else
			{
				printf("Error: unknown BSP version %li\n", bspheader.version);
				flag_error = 2;
				break;
			}

			/* get memory for entity list */
			entlist = malloc(filelength);
			if (!entlist)
			{
				printf("Error: couldn't allocate memory for entity list\n");
				flag_error = 2;
				break;
			}
			memset(entlist, 0, filelength);

			/* read entity list */
			if ( fseek(data, fileoffset, SEEK_SET) || ftell(data) != fileoffset )
			{
				printf("Error: couldn't seek bsp entity list - %s\n", strerror(errno));
				flag_error = 2;
				break;
			}
			if ( fread(entlist, filelength, 1, data) != 1 )
			{
				printf("Error: couldn't read bsp entity list - %s\n", strerror(errno));
				flag_error = 2;
				break;
			}
		}
		else
		{
			/* open map file */
			data = fopen(data_filename, "r");
			if (!data)
			{
				printf("Error: could not open map file \"%s\" - %s\n", data_filename, strerror(errno));
				flag_error = 2;
				break;
			}

			/* get file length */
			if ( fseek(data, 0, SEEK_END) != 0)
			{
				printf("Error: couldn't seek file end - %s\n", strerror(errno));
				flag_error = 2;
				break;
			}
			filelength = ftell(data);
			if ( filelength < 0)
			{
				printf("Error: couldn't get file length - %s\n", strerror(errno));
				flag_error = 2;
				break;
			}
			if ( fseek(data, 0, SEEK_SET) != 0)
			{
				printf("Error: couldn't seek file start - %s\n", strerror(errno));
				flag_error = 2;
				break;
			}

			/* get memory for entity list */
			entlist = malloc(filelength+1);
			if (!entlist)
			{
				printf("Error: couldn't allocate memory for entity list\n");
				flag_error = 2;
				break;
			}
			memset(entlist, 0, filelength+1);

			entlist_char = entlist;
			while (!feof(data) && fgets(entlist_char, filelength, data))
			{
				filelength -= strlen(entlist_char);
				entlist_char += strlen(entlist_char);
			}
		}

		if (entlist)
		{
			/* open ent file */
			ent = fopen(ent_filename, "w");
			if (!ent)
			{
				printf("Error: could not create ent file \"%s\" - %s\n", ent_filename, strerror(errno));
				flag_error = 2;
				break;
			}

			/* write entity list ... */
			if (!flag_invert)
			{
				/* ... as is */
				if ( fputs(entlist, ent) == EOF )
				{
					printf("Error: couldn't write entity list - %s\n", strerror(errno));
					flag_error = 2;
					break;
				}
			}
			else
			{
				/* ... inverted */
				entlist_char = entlist;
				linecount = 0;
				level = 0;
				attrlist = NULL;

				while ( get_line() )
				{
					if (flag_error)
					{
						break;
					}

					get_firstchar();

					if ((*line_char == '{') || (*line_char == '}'))
					{
						if ((level == 1) && (attrlist))
						{
							while (attrlist)
							{
								attribute = attrlist;
								if ( fputs(attribute->line, ent) == EOF )
								{
									printf("Error: couldn't write entity list - %s\n", strerror(errno));
									flag_error = 2;
									break;
								}
								attrlist = attribute->next;
								free(attribute);
							}
						}

						if ( fputs(line, ent) == EOF )
						{
							printf("Error: couldn't write entity list - %s\n", strerror(errno));
							flag_error = 2;
							break;
						}

						if (*line_char == '{')
						{
							level++;
						}
						else
						{
							level--;
						}
					}
					else
					{
						if (level != 1)
						{
							if ( fputs(line, ent) == EOF )
							{
								printf("Error: couldn't write entity list - %s\n", strerror(errno));
								flag_error = 2;
								break;
							}
						}
						else
						{
							attribute = malloc(sizeof(attrentry));
							if (!attribute)
							{
								printf("Error: couldn't allocate memory for entity list\n");
								flag_error = 2;
								break;
							}
							strcpy(attribute->line, line);
							attribute->next = attrlist;
							attrlist = attribute;
						}
					}
				}

				/* write all lines if error occured or data file ended without closing entity */
				while (attrlist)
				{
					attribute = attrlist;
					if ( fputs(attribute->line, ent) == EOF )
					{
						printf("Error: couldn't write entity list - %s\n", strerror(errno));
						flag_error = 2;
						break;
					}
					attrlist = attribute->next;
					free(attribute);
				}

				if (flag_error)
				{
					break;
				}

				linecount--;
			}
		}
	} while (0);

	if (entlist)
	{
		free(entlist);
	}

	if (data)
	{
		fclose(data);
		data = NULL;
	}

	if (ent)
	{
		fclose(ent);
		ent = NULL;
	}
}

/*
******************************************************************************
*
*   M A I N   F U N C T I O N   ( S T A R T   O F   P R O G R A M )
*
******************************************************************************
*/

/*
==============================================================================
here the program starts and all sub functions are called
==============================================================================
*/
int main(int argc, char *argv[])
{
	/* initialize variables */
	flag_error = 0;
	flag_warning = 0;
	flag_invert = 0;
	*data_filename = 0;
	*ent_filename = 0;

	/* display program header lines */
	printf("BSP Entity Extractor\n");
	printf("(c) 1999 by Matthias \"Maddes\" Buecher\n");
	printf("Version %s\n\n", VERSION);

	/* check parameters */
	chk_param(argc, argv);
	if (flag_error)
	{
		exit(flag_error);
	}

	/* here we go with the "real" program */
	process_file();
	if (flag_error)
	{
		exit(flag_error);
	}

	printf("Finished\n");

	return 0;
}
